import logging

from twisted.internet.defer import inlineCallbacks, returnValue, succeed

from ensemble.errors import ProviderError
from ensemble.providers.common.base import MachineProviderBase

from .cobbler import CobblerClient
from .files import FileStorage
from .launch import OrchestraLaunchMachine
from .machine import machine_from_dict, OrchestraMachine

log = logging.getLogger("ensemble.orchestra")


class MachineProvider(MachineProviderBase):

    def __init__(self, environment_name, config):
        super(MachineProvider, self).__init__(environment_name, config)
        self.cobbler = CobblerClient(config)

    def get_file_storage(self):
        """Return a WebDAV-backed FileStorage abstraction."""
        if "storage-url" not in self.config:
            return FileStorage(
                "http://%(orchestra-server)s/webdav" % self.config)
        return FileStorage(self.config["storage-url"])

    def start_machine(self, machine_data, master=False):
        """Start a machine in the provider.

        @param machine_data: a dictionary of data to pass along to the newly
           launched machine.

        @param master: if True, machine will initialize the ensemble admin
            and run a provisioning agent.
        """
        return OrchestraLaunchMachine(self, master).run(machine_data)

    @inlineCallbacks
    def get_machines(self, instance_ids=()):
        """List machines running in the provider.

        @param instance_ids: ids of instances you want to get. Leave empty
            to list all machines owned by this provider.

        @return: a list of OrchestraMachines.

        @raise: MachinesNotFound
        """
        instances = yield self.cobbler.describe_systems(*instance_ids)
        returnValue([machine_from_dict(i) for i in instances])

    @inlineCallbacks
    def shutdown_machines(self, machines):
        """Terminate machines associated with this provider.

        @param machines: list of machines to shut down
        """
        if not machines:
            returnValue([])

        for machine in machines:
            if not isinstance(machine, OrchestraMachine):
                raise ProviderError("Can only shut down OrchestraMachines; "
                                    "got a %r" % type(machine))

        ids = [m.instance_id for m in machines]
        killable_machines = yield self.get_machines(ids)
        for machine in killable_machines:
            yield self.cobbler.stop_system(machine.instance_id)
            yield self.cobbler.release_system(machine.instance_id)
        returnValue(killable_machines)

    def open_port(self, machine, machine_id, port, protocol="tcp"):
        """Authorizes `port` using `protocol` for `machine`."""
        log.warn("Firewalling is not yet implemented in Orchestra")
        return succeed(None)

    def close_port(self, machine, machine_id, port, protocol="tcp"):
        """Revokes `port` using `protocol` for `machine`."""
        log.warn("Firewalling is not yet implemented in Orchestra")
        return succeed(None)

    def get_opened_ports(self, machine, machine_id):
        """Returns a set of open (port, protocol) pairs for `machine`."""
        log.warn("Firewalling is not yet implemented in Orchestra")
        return succeed(None)
