from functools import partial
import os
import shutil

from twisted.internet.defer import inlineCallbacks, returnValue, succeed, fail
from twisted.web.error import Error
from twisted.web.client import downloadPage

from ensemble.formula import get_formula_from_path
from ensemble.formula.bundle import FormulaBundle
from ensemble.errors import FileNotFound
from ensemble.formula.publisher import FormulaPublisher
from ensemble.formula.tests.test_directory import sample_directory

from ensemble.environment.config import EnvironmentsConfig
from ensemble.state.environment import EnvironmentStateManager
from ensemble.state.errors import FormulaStateNotFound
from ensemble.state.tests.common import StateTestBase

from ensemble.unit.formula import download_formula

from ensemble.lib.mocker import MATCH

SAMPLE_ENV = """\
ensemble: environments

environments:
  myfirstenv:
    type: dummy
    foo: bar
    storage-directory: %s
"""


class FormulaPublisherTestBase(StateTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(FormulaPublisherTestBase, self).setUp()

        config = self.get_test_environment_config()
        environment = config.get_default()

        # Store the environment to zookeeper
        environment_state_manager = EnvironmentStateManager(self.client)
        yield environment_state_manager.set_config_state(config, "myfirstenv")

        # Load the environment
        self.provider = environment.get_machine_provider()
        self.storage = self.provider.get_file_storage()

    def get_test_environment_config(self):
        sample_config = SAMPLE_ENV % self.makeDir()
        config = EnvironmentsConfig()
        config.parse(sample_config)
        return config

    @inlineCallbacks
    def publish_formula(self, formula_path=sample_directory):
        formula = get_formula_from_path(formula_path)
        publisher = FormulaPublisher(self.client, self.storage)
        yield publisher.add_formula("local", formula)
        formula_states = yield publisher.publish()
        returnValue((formula, formula_states[0]))


class DownloadTestCase(FormulaPublisherTestBase):

    @inlineCallbacks
    def test_formula_download_file(self):
        """Downloading a formula should store the formula locally.
        """
        formula, formula_state = yield self.publish_formula()
        formula_directory = self.makeDir()

        # Download the formula
        yield download_formula(
            self.client, formula_state.id, formula_directory)

        # Verify the downloaded copy
        checksum = formula.get_sha256()
        formula_key = "%s:%s" % (
            formula.metadata.get_formula_id("local"), checksum)
        formula_path = os.path.join(formula_directory, formula_key)

        self.assertTrue(os.path.exists(formula_path))
        bundle = FormulaBundle(formula_path)
        self.assertTrue(
            bundle.metadata.get_formula_id("local"),
            formula.metadata.get_formula_id("local"))

        self.assertEqual(checksum, bundle.get_sha256())

    @inlineCallbacks
    def test_formula_missing_download_file(self):
        """Downloading a file that doesn't exist raises FileNotFound.
        """
        formula, formula_state = yield self.publish_formula()
        formula_directory = self.makeDir()

        # Delete the file
        file_path = formula_state.bundle_url[len("file://"):]
        os.remove(file_path)

        # Download the formula
        yield self.assertFailure(
            download_formula(self.client, formula_state.id, formula_directory),
            FileNotFound)

    @inlineCallbacks
    def test_formula_download_http(self):
        """Downloading a formula should store the formula locally.
        """
        mock_storage = self.mocker.patch(self.storage)

        def match_string(expected, value):
            self.assertTrue(isinstance(value, basestring))
            self.assertIn(expected, value)
            return True

        mock_storage.get_url(MATCH(
            partial(match_string, "formulas/local:dummy-1")))

        self.mocker.result("http://example.com/foobar.zip")

        download_page = self.mocker.replace(downloadPage)
        download_page(
            MATCH(partial(match_string, "http://example.com/foobar.zip")),
            MATCH(partial(match_string, "local:dummy-1")))

        def bundle_in_place(url, local_path):
            # must keep ref to formula else temp file goes out of scope.
            formula = get_formula_from_path(sample_directory)
            bundle = formula.as_bundle()
            shutil.copyfile(bundle.path, local_path)

        self.mocker.call(bundle_in_place)
        self.mocker.result(succeed(True))
        self.mocker.replay()

        formula, formula_state = yield self.publish_formula()
        formula_directory = self.makeDir()
        self.assertEqual(
            formula_state.bundle_url, "http://example.com/foobar.zip")

        # Download the formula
        yield download_formula(
            self.client, formula_state.id, formula_directory)

    @inlineCallbacks
    def test_formula_download_http_error(self):
        """Errors in donwloading a formula are reported as formula not found.
        """
        def match_string(expected, value):
            self.assertTrue(isinstance(value, basestring))
            self.assertIn(expected, value)
            return True

        mock_storage = self.mocker.patch(self.storage)
        mock_storage.get_url(
            MATCH(partial(match_string, "formulas/local:dummy-1")))
        remote_url = "http://example.com/foobar.zip"
        self.mocker.result(remote_url)

        download_page = self.mocker.replace(downloadPage)
        download_page(
            MATCH(partial(match_string, "http://example.com/foobar.zip")),
            MATCH(partial(match_string, "local:dummy-1")))

        self.mocker.result(fail(Error("400", "Bad Stuff", "")))
        self.mocker.replay()

        formula, formula_state = yield self.publish_formula()
        formula_directory = self.makeDir()
        self.assertEqual(formula_state.bundle_url, remote_url)

        error = yield self.assertFailure(
            download_formula(self.client, formula_state.id, formula_directory),
            FileNotFound)
        self.assertIn(remote_url, str(error))

    @inlineCallbacks
    def test_formula_download_not_found(self):
        """An error is raised if trying to download a non existant formula.
        """
        formula_directory = self.makeDir()

        # Download the formula
        error = yield self.assertFailure(
            download_formula(
                self.client, "local:mickey-21", formula_directory),
            FormulaStateNotFound)

        self.assertEquals(str(error),
                          "Formula 'local:mickey-21' was not found")
